#include <iostream>
#include <iterator>
#include <iomanip>
#include <fstream>
#include <string>
#include <ctime>
#include <cstdlib>
#include <utility>
#include <algorithm>  // for min
using namespace std;

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#include "mutualfund.h"

void MutualFund::readMutualFundHistoryFile(const string& filename)
{
    string buffer;
    string tempNavString;
    Date tempDate;
    float tempNav;
    size_t loc4thComma;
    size_t loc5thComma;
    size_t count = 0; // record count from file
    cout << "Reading file " << filename << " into map<Date,float> ...";
    ifstream hisFin(filename);
    if (!hisFin)
    {
        cerr << "Unable to open history file: " << filename << endl;
        exit(1);
    }

    while (getline(hisFin, buffer))
    {
        tempDate = Date(buffer.substr(0, 10));
        loc4thComma = locationOfNthCommaInString(buffer, 4);
        loc5thComma = locationOfNthCommaInString(buffer, 5);
        tempNavString = buffer.substr(loc4thComma + 1, loc5thComma - loc4thComma);
        tempNav = stof(tempNavString);
        navHistory[Date(tempDate)] = tempNav;
        count++;
    }
    cout << endl << count << " records read" << endl << endl;
}

void MutualFund::readMutualFundDividendFile(const string& filename)
{
    string buffer;
    Date tempDate; // today date
    float tempDiv;
    size_t count = 0; // record count from file
    cout << "Reading file " << filename << " into set<Dividend> ...";

    ifstream divFin(filename.c_str());
    if (!divFin)
    {
        cerr << "Unable to open dividend file: " << filename << endl;
        exit(1);
    }

    getline(divFin, buffer); // skip first line
    while (divFin)
    {
        buffer.clear();
        getline(divFin, buffer); // read record from input file
        if (buffer.size() < 3)
        {
            break;
        }
        tempDate = Date(buffer.substr(0, 10));
        tempDiv = stof(buffer.substr(11));
        divHistory.insert(Dividend(tempDate, tempDiv));
        count++;
    }
    cout << endl << count << " records read" << endl << endl;
}

size_t locationOfNthCommaInString(const string& s, unsigned short whichOne)
{
    if (whichOne < 1)
        return string::npos;
    size_t pos = 0;
    for (unsigned short wh = 0; wh < whichOne; wh++)
    {
        pos = s.find(',', pos + 1);
    }
    return pos;
}


Dividend::Dividend(const Date& d, float amount)
    : date(d), div(amount)
{
}

std::ostream& operator<<(ostream& out, const Dividend& obj)
{
    out << obj.get_date() << "   " << obj.get_div();
    return out;
}

ostream& operator<<(ostream& out, const std::set<Dividend>& obj)
{
    copy(obj.cbegin(), obj.cend(),ostream_iterator<Dividend>(out,"\n"));
    return out;
}

ostream& operator<<(ostream& out, const std::pair<Date, float>& obj)
{
    out << setw(10) << obj.first << setw(10) << obj.second;
    return out;
}

ostream& operator<<(ostream& out, const std::map<Date, float>& obj)
{
    copy(obj.cbegin(), obj.cend(),ostream_iterator<std::pair<Date, float>>(out,"\n"));
    return out;
}

MutualFund::MutualFund(const string& tick)
    : ticker(tick)
{
    string MF_HistoryFile;
    string MF_DividendFile;
    readMutualFundHistoryFile(Path + tick + "shuffled.csv");
    readMutualFundDividendFile(Path + tick + "div.csv");
    oldestDate = navHistory.cbegin()->first;
    newestDate = navHistory.crbegin()->first;
    newestNav = navHistory.crbegin()->second;
    writeNavDataToFile();
    writeDivDataToFile();
}

void MutualFund::writeNavDataToFile() const
{
    string filename = Path + ticker + "nav.txt";
    ofstream fout(filename);
    if (!fout)
    {
        cerr << "Unable to open output file " << filename << "\nExiting ... " << endl;
        exit(EXIT_FAILURE);
    }
    cout << "Writing NAV history data into file " << filename << " ... ";
    fout << setprecision(2) << fixed << navHistory << endl;
    cout << "done" << endl;
}

void MutualFund::writeDivDataToFile() const
{
    string filename = Path + ticker + "div.txt";
    ofstream fout(filename);
    if (!fout)
    {
        cerr << "Unable to open output file " << filename << "\nExiting ... " << endl;
        exit(EXIT_FAILURE);
    }
    cout << "Writing dividend history data into file " << filename << " ... ";
    fout << setprecision(3) << fixed << divHistory << endl;

     cout << "done" << endl << endl;
}

float MutualFund::getMovingAverage(int NumDays) const
{
    float sum = 0;
    auto it = navHistory.rbegin();
    for (auto i = 0; i < NumDays; ++i, ++it)
    {
        sum += it->second;
    }
    return sum / NumDays;
}

pair<float, float> MutualFund::getMinMaxForTheLastYear() const
{
    time_t temp_time_t = newestDate.get_value();
    tm temp_tm = *localtime(&temp_time_t);
    temp_tm.tm_year -= 1;
    Date oneYearAgo(mktime(&temp_tm));
    auto startIt = navHistory.upper_bound(oneYearAgo);
    auto endIt = navHistory.cend();
    auto minValue = min_element(startIt, endIt, navCompare)->second;
    auto maxValue = max_element(startIt, endIt, navCompare)->second;
    return pair<float, float>(minValue, maxValue);
}

float MutualFund::ThreeYearGain() const
{
    time_t temp_time_t = newestDate.get_value();
    tm temp_tm = *localtime(&temp_time_t);
    temp_tm.tm_year -= 3;
    Date threeYearsAgo(mktime(&temp_tm));
    return getReturnOverTime(InitialInvestment, threeYearsAgo, newestDate);
}


bool navCompare(const pair<Date, float>& a, const pair<Date, float>& b)
{
    return a.second < b.second;
}

map<Date, float>::const_iterator MutualFund::getNthNavFromNavHistory(int n) const
{
    auto it = navHistory.begin();
    for (auto i = 0; i < n; ++i)
        ++it;
    return it;
}

pair<map<Date, float>::const_iterator, map<Date, float>::const_iterator> MutualFund::getTwoRandomNavs() const
{
    Date date1, date2;
    int rand1 = 0, rand2 = 0;
    do
    {
        rand1 = rand() % (navHistory.size());
        rand2 = rand() % (navHistory.size());
    }
    while (rand2 - rand1 <= 250);
    return pair<map<Date, float>::const_iterator, map<Date, float>::const_iterator>(getNthNavFromNavHistory(rand1), getNthNavFromNavHistory(rand2));
}

void MutualFund::report() const
{
    cout << setprecision(2) << fixed;
    cout << "Fund: " << ticker << endl;
    cout << "Analysis Date: " << newestDate << endl;
    cout << "Latest NAV: $" << newestNav << endl;
    auto minmax = getMinMaxForTheLastYear();
    cout << "Minimum for the last year = $" << minmax.first << endl;
    cout << "Maximum for the last year = $" << minmax.second << endl;
    cout << "50 day moving average = $" << getMovingAverage(50) << endl;
    float movingAverage200 = getMovingAverage(200);
    cout << "200 day moving average = $" << movingAverage200 << endl;
    cout << "% gain over 200 Day Moving Avg: " << 100 * (newestNav - movingAverage200) / movingAverage200 << '%' << endl;
    float gain = ThreeYearGain() - InitialInvestment;
    cout << "% return over the last 3 years: " << 100 * gain / InitialInvestment << '%' << endl << endl;
}


float MutualFund::getReturnOverTime(float initialInvestment, const Date& start, const Date& end) const
{
    Date copyOfStart(start);
    // Three tries to find a good start date
    if (navHistory.find(copyOfStart) == navHistory.end())
    {
        ++copyOfStart;
        if (navHistory.find(copyOfStart) == navHistory.end())
        {
            ++copyOfStart;
            if (navHistory.find(copyOfStart) == navHistory.end())
            {
                ++copyOfStart;
            }
        }
    }
    float purchasePrice = navHistory.lower_bound(copyOfStart)->second;
    float sharesOwned = initialInvestment / purchasePrice;
    sharesOwned = applyDividends(sharesOwned, copyOfStart, end);

    float sellingPrice = navHistory.lower_bound(end)->second;
    float endValue = sharesOwned * sellingPrice;
    return endValue;
}

// Returns the number of shares owned after applying dividends
float MutualFund::applyDividends(float sharesOwned,const Date& start,const Date& end) const
{
    float dividend;
    float amountOfDividendPayment;
    Date dividendDate;
    float sharePriceOnDividendDate;
    float sharesPurchased;
    string divFilename = "C:/temp/dividend_detail.txt";
    static ofstream fout(divFilename.c_str());
    fout << fixed << setprecision(2);
    fout << "Dividends for " << ticker << ": " << start << '-' << end << endl;
    fout << "Date      Dividend   Shares    Div Amt  Sh Price Purchased  New Sh Owned" << endl;
    // get iterators for range of dividends
    auto startIt = get_divHistory().lower_bound(start + 30);
    auto endIt = get_divHistory().upper_bound(end);
    for (auto it = startIt; it != endIt; ++it)
    {
        dividend = it->get_div();
        dividendDate = it -> get_date();
        amountOfDividendPayment = sharesOwned * dividend;
        // get share price on dividend date
        sharePriceOnDividendDate = navHistory.find(dividendDate)->second;
        sharesPurchased = amountOfDividendPayment / sharePriceOnDividendDate;
        fout << dividendDate << setw(8) << setprecision(3) << dividend << setw(10) << sharesOwned;
        sharesOwned += sharesPurchased;
        fout << setw(10) << setprecision(2) << amountOfDividendPayment << setw(10) << sharePriceOnDividendDate
             << setw(9) << setprecision(3) << sharesPurchased << setw(12) << sharesOwned << endl;
    }
    fout << endl;
    return sharesOwned;
}
